@@ -200,7 +200,7 @@ class FSO(object):
     '''
     def __init__(self):
         self.wc_path = None
-        self.state = [ "-", "-" ] # '-': absent, 'F': file, 'D': dir
+        self.state = [ "-", "-" ] # '-': absent, 'F': file, 'D': dir, 'L': symlink
 
     def status(self):
         return "[%s%s]" % (self.state[S_WC], self.state[S_IM])
@@ -216,8 +216,8 @@ class FSOCollection(dict):
     Collection of FSOs
     '''
     def print(self):
-        print(" / Status in working copy (-:absent, F:file, D:dir)")
-        print(" |/ Status in imported sources (-:absent, F:file, D:dir)")
+        print(" / Status in working copy (-:absent, F:file, D:dir, L:link)")
+        print(" |/ Status in imported sources (-:absent, F:file, D:dir, L:link)")
         for k in sorted(self.keys(), key=filename_sort_key):
             e = self[k]
             print("%s %s%s" % (e.status(), shlex.quote(k),
@@ -310,9 +310,10 @@ class Config(dict):
     def print(self):
         for k in sorted(self):
             o = self[k]
-            print("# %s" % o.helpmsg)
+            for s in o.helpmsg.split('\n'):
+                print("# %s" % s)
             print("%-20s: %s" % (k, str(o)))
-            print()
+            print("")
 
 
 class SvnVndImport(cmd.Cmd):
@@ -330,6 +331,18 @@ class SvnVndImport(cmd.Cmd):
         self.importdir = importdir
         self.svninfo = svninfo
         self.config = Config()
+        self.config.add_option('symlink-handling',
+                ConfigOpt("as-is", "How symbolic links are handled;\n" +
+                    "  'dereference' treats as normal files/dirs (and " +
+                    "ignores dangling links);\n" +
+                    "  'as-is' imports as symlinks"))
+        self.config.add_option('exec-permission',
+                ConfigOpt("preserve", "How 'executable' permission bits " +
+                    "are handled;\n" +
+                    "  'preserve' sets svn:executable property as in " +
+                    "imported sources;\n" +
+                    "  'clear' removes svn:executable on all new files " +
+                    "(but keeps it intact on existing files)."))
         self.config.add_option('save-diff-copied',
                 ConfigOpt(None, "Save 'svn diff' output on the " +
                     "moved/copied files and directories to this " +
@@ -359,12 +372,13 @@ class SvnVndImport(cmd.Cmd):
     def scan(self):
         self.items = FSOCollection()
         self.info(1, "Scanning working copy directory...")
-        self.get_lists(self.wcdir, S_WC)
+        self.get_lists(self.wcdir, S_WC, False)
         self.info(1, "Scanning imported directory...")
-        self.get_lists(self.importdir, S_IM)
+        self.get_lists(self.importdir, S_IM,
+                self.config.get('symlink-handling') == "dereference")
 
-    def get_lists(self, top, where):
-        for d, dn, fn in os.walk(top, followlinks=True):
+    def get_lists(self, top, where, deref):
+        for d, dn, fn in os.walk(top, followlinks=deref):
             dr = os.path.relpath(d, top)
             # If under .svn directory at the top (SVN 1.7+) or has .svn
             # in the path (older SVN), ignore
@@ -374,9 +388,36 @@ class SvnVndImport(cmd.Cmd):
                 continue
             if dr != '.':
                 self.items.add(dr, where, "D")
-            for f in fn:
+            dnn = [] # List where we'll descend
+            for f in fn + dn:
                 fr = os.path.normpath(os.path.join(dr, f))
-                self.items.add(fr, where, "F")
+                frp = os.path.join(d, f)
+                if os.path.islink(frp):
+                    if deref:
+                        # Dereferencing:
+                        # - check for dangling/absolute/out-of-tree symlinks and abort
+                        rl = os.readlink(frp)
+                        if not os.path.exists(frp):
+                            self.info(1, "WARN: Ignoring dangling symlink %s -> %s" % (fr, rl))
+                            continue
+                        if os.path.isabs(rl):
+                            self.info(1, "WARN: Ignoring absolute symlink %s -> %s" % (fr, rl))
+                            continue
+                        tgt = os.path.normpath(os.path.join(dr, rl))
+                        if tgt == ".." or tgt.startswith(".." + os.sep):
+                            self.info(1, "WARN: Ignoring out-of-wc symlink %s -> %s" % (fr, rl))
+                            continue
+                    else:
+                        # Importing symlinks as-is, no need to check.
+                        self.items.add(fr, where, "L")
+                        continue
+                # If we get here, treat symlinks to files as regular files, and add directories
+                # to the list of traversed subdirs
+                if os.path.isfile(frp):
+                    self.items.add(fr, where, "F")
+                if os.path.isdir(frp):
+                    dnn.append(f)
+            dn[:] = dnn
 
     def onecmd(self, str):
         'Override for checking number of arguments'
@@ -385,7 +426,7 @@ class SvnVndImport(cmd.Cmd):
         except InvalidUsageException as e:
             if e.cmd is not None:
                 print("!!! Invalid usage of `%s' command: %s" % (e.cmd, e))
-                print()
+                print("")
                 self.onecmd("help " + e.cmd)
             else:
                 print("!!! %s" % e)
@@ -403,6 +444,10 @@ class SvnVndImport(cmd.Cmd):
         pos = 0
         atatime = 100
         output = ""
+        # svn treats '@' specially (peg revision); if there's such character in a
+        # file name - append an empty peg revision
+        args_fixed = list(map(lambda x: x + "@" if x.find("@") != -1 else x, args_fixed))
+        args_split = list(map(lambda x: x + "@" if x.find("@") != -1 else x, args_split))
         while pos < len(args_split) or (pos == 0 and len(args_split) == 0):
             svnargs = ['svn'] + args_fixed + args_split[pos : pos + atatime]
             pos += atatime
@@ -558,6 +603,10 @@ class SvnVndImport(cmd.Cmd):
         if xsrc.state[S_WC] != xdst.state[S_IM]:
             # Different kinds - definitely not the same object
             return 0
+        if xsrc.state[S_WC] == "L" or xdst.state[S_IM] == "L":
+            # Symlinks are not considered the same object (same target in
+            # different dirs refers to different objects).
+            return 0
         if xsrc.state[S_WC] == "D":
             return self.similarity_dir(src, dst, threshold, lst_removal)
         else:
@@ -734,8 +783,10 @@ class SvnVndImport(cmd.Cmd):
                  edit and load.
         '''
         self.parse_args(arg, 0, "detect")
+        # Configurable for file/dirs; symlinks are never similar.
         self.detect({ "D": self.config.get('dir-similarity'),
-            "F": self.config.get('file-similarity')})
+            "F": self.config.get('file-similarity'),
+            "L": 1001 })
 
     def do_apply(self, arg):
         '''
@@ -763,7 +814,28 @@ class SvnVndImport(cmd.Cmd):
         dirs_removed = []
         files_added = []
         files_removed = []
+        files_set_exec = []
+        files_clear_exec = []
+
         self.info(2, "  Creating dirs and copying files...")
+        def copyfile_helper(i, nk_wc):
+            '''Helper: copy a file and optionally, transfer permissions.'''
+            f = os.path.join(self.importdir, i)
+            t = os.path.join(self.wcdir, i)
+            shutil.copyfile(f, t)
+            # If exec-permission is 'clear', we don't need to do anything:
+            # shutil.copyfile will create the file as non-executable.
+            if self.config.get('exec-permission') == 'preserve':
+                # If the file is new, just copying the mode is enough:
+                # svn will set the svn:executable upon adding it.
+                if nk_wc == "F":
+                    # Existing file, check what the setting shall be
+                    if os.access(f, os.X_OK) and not os.access(t, os.X_OK):
+                        files_set_exec.append(i)
+                    elif not os.access(f, os.X_OK) and os.access(t, os.X_OK):
+                        files_clear_exec.append(i)
+                shutil.copymode(f, t)
+
         for i in sorted(self.items.keys()):
             e = self.items[i]
             nk_wc = e.state[S_WC]
@@ -778,14 +850,46 @@ class SvnVndImport(cmd.Cmd):
                     flg = "(added dir)"
                 elif nk_im == "F":
                     # New file added
-                    shutil.copyfile(os.path.join(self.importdir, i),
-                            os.path.join(self.wcdir, i))
+                    copyfile_helper(i, nk_wc);
                     files_added.append(i)
                     flg = "(added file)"
+                elif nk_im == "L":
+                    tim = os.readlink(os.path.join(self.importdir, i))
+                    os.symlink(tim, os.path.join(self.wcdir, i))
+                    files_added.append(i)
+                    flg = "(added symlink)"
                 else:
                     # Not in imported sources, not in WC (moved
                     # away/removed) - nothing to do
                     pass
+            elif nk_wc == "L":
+                # Symbolic link in a working copy
+                if nk_im == "L":
+                    # Symbolic link in both. If the same target, do nothing. Otherwise,
+                    # replace.
+                    twc = os.readlink(os.path.join(self.wcdir, i))
+                    tim = os.readlink(os.path.join(self.importdir, i))
+                    if tim != twc:
+                        self.run_svn(["rm", "--force", i])
+                        os.symlink(tim, os.path.join(self.wcdir, i))
+                        files_added.append(i)
+                        flg = "(replaced symlink)"
+                elif nk_im == "D":
+                    # Was a symlink, now a directory. Replace.
+                    self.run_svn(["rm", "--force", i])
+                    os.mkdir(os.path.join(self.wcdir, i))
+                    dirs_added.append(i)
+                    flg = "(replaced symlink with dir)"
+                elif nk_im == "F":
+                    # Symlink replaced with file.
+                    self.run_svn(["rm", "--force", i])
+                    copyfile_helper(i, nk_wc);
+                    files_added.append(i)
+                    flg = "(replaced symlink with file)"
+                else:
+                    # Was a symlink, removed
+                    files_removed.append(i)
+                    flg = "(removed symlink)"
             elif nk_wc == "F":
                 # File in a working copy
                 if nk_im == "D":
@@ -796,9 +900,15 @@ class SvnVndImport(cmd.Cmd):
                     flg = "(replaced file with dir)"
                 elif nk_im == "F":
                     # Was a file, is a file - just copy contents
-                    shutil.copyfile(os.path.join(self.importdir, i),
-                            os.path.join(self.wcdir, i))
+                    copyfile_helper(i, nk_wc);
                     flg = "(copied)"
+                elif nk_im == "L":
+                    # Was a file, now a symlink. Replace.
+                    self.run_svn(["rm", "--force", i])
+                    tim = os.readlink(os.path.join(self.importdir, i))
+                    os.symlink(tim, os.path.join(self.wcdir, i))
+                    files_added.append(i)
+                    flg = "(replaced file with symlink)"
                 else:
                     # Was a file, removed
                     files_removed.append(i)
@@ -811,12 +921,21 @@ class SvnVndImport(cmd.Cmd):
                 elif nk_im == "F":
                     # Directory replaced with file. Need to remove dir
                     # immediately, as bulk removals/additions assume new files
-                    # and dirs already in place.
+                    # and dirs already in place. Also, removing a directory
+                    # removes all its descendants - mark them as removed.
                     self.run_svn(["rm", "--force", i])
-                    shutil.copyfile(os.path.join(self.importdir, i),
-                            os.path.join(self.wcdir, i))
+                    self.items.wc_remove(i)
+                    copyfile_helper(i, nk_wc);
                     files_added.append(i)
                     flg = "(replaced dir with file)"
+                elif nk_im == "L":
+                    # Was a directory, now a symlink. Replace.
+                    self.run_svn(["rm", "--force", i])
+                    self.items.wc_remove(i)
+                    tim = os.readlink(os.path.join(self.importdir, i))
+                    os.symlink(tim, os.path.join(self.wcdir, i))
+                    files_added.append(i)
+                    flg = "(replaced dir with symlink)"
                 else:
                     # Directory removed
                     dirs_removed.append(i)
@@ -832,7 +951,7 @@ class SvnVndImport(cmd.Cmd):
             dirs_added, files_added))
         dirs_added = list(filter(lambda x: os.path.dirname(x) not in
             dirs_added, dirs_added))
-        self.info(2, "  Running SVN add/rm commands");
+        self.info(2, "  Running SVN add/rm/propset/propdel commands");
         if len(dirs_added):
             self.run_svn(["add"], dirs_added)
         if len(files_added):
@@ -841,6 +960,10 @@ class SvnVndImport(cmd.Cmd):
             self.run_svn(["rm"], dirs_removed)
         if len(files_removed):
             self.run_svn(["rm"], files_removed)
+        if len(files_set_exec):
+            self.run_svn(["propset", "svn:executable", "*"], files_set_exec)
+        if len(files_clear_exec):
+            self.run_svn(["propdel", "svn:executable"], files_clear_exec)
         # Save the diff for the copied/moved items
         diff_save = self.config.get('save-diff-copied')
         if diff_save is not None:
@@ -918,7 +1041,7 @@ class SvnVndImport(cmd.Cmd):
         colsz = int((self.termwidth - 14) / 2)
         if len(self.prepare_ops):
             print("Currently recorded preparatory operations:")
-            print()
+            print("")
             print("%5s  %s  %-*s  %-*s" %
                     ("#", "Op", colsz, "Source", colsz, "Destination"))
             for id, o in enumerate(self.prepare_ops):
@@ -930,10 +1053,10 @@ class SvnVndImport(cmd.Cmd):
                             (id, o[0], colsz, o[1], colsz, o[2]))
                 else:
                     print("%5d  %s  %-*s" % (id, o[0], colsz, o[1]))
-            print()
+            print("")
         else:
             print("No copies/moves/removals recorded")
-            print()
+            print("")
 
     def do_save(self, arg):
         '''
@@ -1039,7 +1162,7 @@ if __name__ == '__main__':
     if p.returncode != 0:
         print("%s: does not appear to be SVN working copy." % args.wcdir)
         print("`svn info' exited with status %d and returned:" % p.returncode)
-        print()
+        print("")
         print(se.decode())
         sys.exit(1)
     imp = SvnVndImport(args.wcdir, args.importdir, so.decode())
